import 'dart:io';

import 'package:cron/cron.dart';
import 'package:dshell/dshell.dart';
import 'package:meta/meta.dart';
import 'package:path/path.dart';

import 'dns_auth.dart';
import 'http_auth.dart';

class Certbot {
  static const LETSENCRYPT_ROOT = '/etc/letsencrypt';

  static const NGINX_CERT_ROOT = '/etc/nginx/certs/';
  static const NGINX_CERT_ROOT_OVERWRITE = 'NGINX_CERT_ROOT_OVERWRITE';

  /// CERTIFICATE_PATH MUST be on a persistent volume so we don't loose the
  /// certificates each time we restart nginx.
  static const CERTIFICATE_PATH = 'config/live';

  /// The file containing the concatenated certs.
  static const CERTIFICATES_FILE = 'fullchain.pem';

  /// our private key.
  static const PRIVATE_KEY_FILE = 'privkey.pem';

  /// The name of the logfile that cerbot writes to.
  /// We also write our log messages to this file.
  static const LOG_FILE_NAME = 'letsencrypt.log';

  static const LIVE_WWW_PATH = '/etc/nginx/live';

  static final Certbot _self = Certbot._internal();
  factory Certbot() => _self;

  /// The certbot log file
  String get logfile => join(letsEncryptLogPath, LOG_FILE_NAME);
  Certbot._internal() {
    print('Logging to $logfile');

    if (!exists(letsEncryptLogPath)) {
      createDir(letsEncryptLogPath, recursive: true);
    }

    if (!exists(letsEncryptWorkPath)) {
      createDir(letsEncryptWorkPath, recursive: true);
    }

    if (!exists(letsEncryptConfigPath)) {
      createDir(letsEncryptConfigPath, recursive: true);
    }
    // logfile = env('LOG_FILE');

    // ArgumentError.checkNotNull(logfile, 'The environment variable "LOG_FILE" must be set to the path of the logfile');
  }

  /// Check that we have valid certificates and deploys them to nginx.
  ///
  /// If we have no certificates then we force nginx into acquire mode.
  ///
  /// The certificates are stored in a persistant volume called 'certificates'
  /// and we need to copy them into /etc/nginx/certs on each start
  /// so that nginx has access to them.
  ///
  void deployCertificates({@required String fqdn}) {
    var hasValidCerts = false;
    if (!exists(getCertificateFullChainPath(fqdn))) {
      printerr(
          "ERROR: The Certfifcate for $fqdn don't exist. Please run 'certbot'");
    } else {
      if (hasExpired(fqdn)) {
        printerr(
            "ERROR The Certificate for $fqdn has expired. Please run 'start certbot");
      } else {
        hasValidCerts = true;
      }
    }

    if (exists(LIVE_WWW_PATH)) {
      deleteSymlink(LIVE_WWW_PATH);
    }

    if (hasValidCerts) {
      /// symlink the user's custom content.
      symlink('/etc/nginx/custom', LIVE_WWW_PATH);
      _deploy(fqdn);
    } else {
      printerr('*' * 80);
      printerr(
          "ERROR: nginx is starting in 'Certificate Acquisition' mode. In this mode it will only respond to CertBot certificate validation requests.");
      printerr('*' * 80);

      /// symlink in the http configs which only permit certbot access
      symlink('/etc/nginx/acquire', LIVE_WWW_PATH);
      reloadNginx();
    }
  }

  /// Used more for testing, but essentially deletes any existing certificates
  /// and places the system into acquire mode.
  /// Could also be used to play with and remove staging certificates
  void revoke({@required String fqdn, bool staging = false}) {
    var workDir = _createDir(Certbot().letsEncryptWorkPath);
    var logDir = _createDir(Certbot().letsEncryptLogPath);
    var configDir = _createDir(Certbot().letsEncryptConfigPath);

    var cmd = 'certbot revoke'
        ' --cert-path ${join(_latestCertificatePath(fqdn), 'cert.pem')}'
        ' --non-interactive '
        ' --work-dir=$workDir '
        ' --config-dir=$configDir '
        ' --logs-dir=$logDir ';

    if (staging) cmd += ' --staging ';

    cmd.start(
        runInShell: true,
        nothrow: true,
        progress:
            Progress((line) => print(line), stderr: (line) => printerr(line)));

    _delete(fqdn);

    /// delete all of the certificates
    // find('*', root: _latestCertificatePath(fqdn))
    //     .forEach((file) => delete(file));
    /// calling deploy will put nginx back into acquire mode.
    deployCertificates(fqdn: fqdn);
  }

  /// used by revoke to delete certificates after they have been revoked
  /// If we don't do this then the revoked certificates will still be renewed.
  void _delete(String fqdn) {
    var workDir = _createDir(Certbot().letsEncryptWorkPath);
    var logDir = _createDir(Certbot().letsEncryptLogPath);
    var configDir = _createDir(Certbot().letsEncryptConfigPath);

    var cmd = 'certbot delete'
        ' --cert-name $fqdn'
        ' --non-interactive '
        ' --work-dir=$workDir '
        ' --config-dir=$configDir '
        ' --logs-dir=$logDir ';

    cmd.start(
        runInShell: true,
        nothrow: true,
        progress:
            Progress((line) => print(line), stderr: (line) => printerr(line)));
  }

  /// Returns the path where the certificates are stored.
  String getCertificateStoragePath(String fqdn) {
    return join(letsEncryptRootPath, CERTIFICATE_PATH, fqdn);
  }

  static String get letsEncryptRootPath {
    /// allow the root to be over-ridden to make testing easier.
    // ignore: unnecessary_cast
    var root = env('LETSENCRYPT_ROOT') as String;
    if (root == null) {
      return LETSENCRYPT_ROOT;
    } else {
      return root;
    }
  }

  String get letsEncryptWorkPath {
    return join(letsEncryptRootPath, 'work');
  }

  String get letsEncryptLogPath {
    return join(letsEncryptRootPath, 'logs');
  }

  String get letsEncryptConfigPath {
    return join(letsEncryptRootPath, 'config');
  }

  /// The path to the fullchain.pem file.
  String getCertificateFullChainPath(String fqdn) {
    return join(getCertificateStoragePath(fqdn), CERTIFICATES_FILE);
  }

  /// The path to the privatekey.pem file.
  String getPrivateKeyPath(String fqdn) {
    return join(getCertificateStoragePath(fqdn), PRIVATE_KEY_FILE);
  }

  /// Checks if the certificate for the given fqdn
  /// has expired
  bool hasExpired(String fqdn) {
    var hasExpired = true;

    var lines = 'certbot certificates '
            ' --work-dir=${Certbot().letsEncryptWorkPath}'
            ' --config-dir=${Certbot().letsEncryptConfigPath}'
            ' --logs-dir=${Certbot().letsEncryptLogPath}'
            ' -d $fqdn'
        .toList(nothrow: true);

    var found = false;
    for (var line in lines) {
      if (line.trim().startsWith('Expiry Date')) {
        found = true;
        var parts = line.trim().split(RegExp(r'\s+'));

        var dateString = parts[2];

        var year = int.parse(dateString.substring(0, 4));
        var month = int.parse(dateString.substring(5, 7));
        var day = int.parse(dateString.substring(8, 10));
        // not working in a unit tests?
        // var format = DateFormat(dateString, 'yyyy-MM-dd');
        // var expiryDate = format.parse(dateString);
        var expiryDate = DateTime(year, month, day);

        if (!expiryDate.isAfter(DateTime.now().subtract(Duration(days: 1)))) ;
        {
          // we are not yet expired
          hasExpired = false;
          break;
        }
      }
    }

    if (!found) {
      throw CertbotException('Unable to find the Expiry Date for $fqdn');
    }

    return hasExpired;
  }

  /// Renews or gets for the first time the certificates
  /// for the given [fqdn]
  /// If [debug] is true then only a staging certificate will be requested.
  void acquire({
    @required String fqdn,
    @required String tld,
    @required String emailaddress,
    @required String mode,
    bool staging = false,
    bool debug = true,
  }) {
    /// first determine the method based on whether we are in public or private mode

    if (mode == 'public') {
      http_auth_renewal(
          fqdn: fqdn,
          tld: tld,
          emailaddress: emailaddress,
          staging: staging,
          debug: debug);
    } else {
      dns_auth_renewal(
          fqdn: fqdn,
          tld: tld,
          emailaddress: emailaddress,
          staging: staging,
          debug: debug);
    }

    _deploy(fqdn);
  }

  void scheduleRenews() {
    var cron = Cron();
    cron.schedule(Schedule.parse('*/3 * * * *'), () async {
      renew();
    });
  }

  void renew() {
    'certbot renew'.run;
  }

  void log(String message) {
    logfile.append(message);
  }

  void logError(String message) {
    logfile.append('*' * 80);
    logfile.append('*');
    logfile.append('*    ERROR: $message');
    logfile.append('*');
    logfile.append('*' * 80);
  }

  void _deploy(String fqdn) {
    var certpath = _latestCertificatePath(fqdn);

    /// we need to leave the original files in place as they form part
    /// of the letsencrypt archive
    copy(join(certpath, 'fullchain.pem'), '/tmp/fullchain.pem',
        overwrite: true);
    copy(join(certpath, 'privkey.pem'), '/tmp/privkey.pem', overwrite: true);

    /// but we need to move them in place using move so that
    /// the replace is essentially atomic so that nginx does see partially
    /// created certificates.
    move('/tmp/fullchain.pem', join(nginxCertPath, 'fullchain.pem'),
        overwrite: true);
    move('/tmp/privkey.pem', join(nginxCertPath, 'privkey.pem'),
        overwrite: true);
    reloadNginx();
  }

  void reloadNginx() {
    if (exists('/var/run/nginx.pid')) {
      /// force nginx to reload its config.
      'nginx -s reload'.run;
    } else {
      Settings().verbose('Nginx reload ignored as nginx is not running');
    }
  }

  static String get nginxCertPath {
    var path = env('NGINX_CERT_ROOT_OVERWRITE');

    path ??= NGINX_CERT_ROOT;
    return path;
  }

  /// Each time certbot creates a new certificate it placess it in a number path
  /// except for the first one.
  /// In order of acquistion
  /// conifg/live/<fqdn>
  /// conifg/live/<fqdn-001>
  /// conifg/live/<fqdn-002>
  String _latestCertificatePath(String fqdn) {
    var livepath = join(Certbot().letsEncryptConfigPath, 'live');
    // if no paths contain '-' then the base fqdn path is correct.
    var latest = join(livepath, fqdn);

    /// find all the dirs that begin with <fqdn> in the live directory.
    var paths =
        find('$fqdn*', root: livepath, types: [FileSystemEntityType.directory])
            .toList();

    var max = 0;
    for (var path in paths) {
      if (path.contains('-')) {
        // noojee.org-0001
        var parts = path.split('-');
        var num = int.parse(parts[1]);
        if (num > max) {
          max = num;
          latest = join(livepath, path);
        }
      }
    }

    return latest;
  }

  String _createDir(String dir) {
    if (!exists(dir)) {
      createDir(dir, recursive: true);
    }
    return dir;
  }
}

class CertbotException implements Exception {
  String message;
  CertbotException(this.message);

  @override
  String toString() => message;
}
